/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.MutexRepository.InstanceLock;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.ReceiverRepository;

/**
 * the runtime behaviour of the pick activity.
 * 
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Pick extends AbstractActivity implements InboundMessageElement {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  protected List<OnMessage> onMessages;
  protected List<OnAlarm> onAlarms;
  /** createInstance attribute of this pick element. */
  protected boolean createInstance = false;
  
  protected Pick() {
    super();
  }

  public Pick(String uuid) {
    super(uuid);
  }

  /**
   * business logic.
   * 
   * @param execution
   *            execution
   */
  public void executeActivity(BpelExecution execution) {
    // add to list of child executions waiting for a signal
    execution.getProcessInstance().addWaitingExecution(execution);
    if (onAlarms != null) {
      for (OnAlarm onAlarm : onAlarms) {
        OnAlarm.createTimer(execution, 
            onAlarm.getDeadlineExpression(), 
            onAlarm.getDurationExpression(),
            onAlarm.getUuid());
      }
    }
    execution.waitForSignal();
  }

  @Override
  public void signal(BpelExecution execution, String signal,
      Map<String, Object> parameters) {
    if (parameters != null) {
      OnMessage onMessage = (OnMessage) parameters.get("receivingElement");
      NodeImpl onAlarmActivity = (NodeImpl) parameters.get("onAlarm");
      if (onMessage != null || onAlarmActivity != null) {
        // Remove other waiting executions
        ReceiverRepository receiverRepository = Environment.getCurrent().get(ReceiverRepository.class);
        QName processQName = execution.getProcessDefinition().getQName();
        for (OnMessage onMsg : onMessages) {
          OperationKey opKey = onMsg.getOperationKey(processQName);
          Receiver receiver = receiverRepository.getReceiver(opKey);
          receiver.removeWaitingExecution(execution);
        }
        // Remove from list of child executions waiting for a signal
        execution.getProcessInstance().getWaitingExecutions().remove(execution);
      }
      if (onMessage != null) {
        NodeImpl node = onMessage.getActivity();
        execution.execute(node);
      } else if (onAlarmActivity != null) {
        execution.execute(onAlarmActivity);
      } else {
        throw new OrchestraException("Unsupported signal parameters");
      }
    } else if (signal != null) {
      boolean found = false;
      for (OnAlarm onAlarm : onAlarms) {
        if (signal.equals(onAlarm.getUuid())) {
          found = true;
          BpelExecution bpelInstance = (BpelExecution) execution.getProcessInstance();
          Map<String, Object> params = new HashMap<String, Object>();
          params.put("onAlarm", onAlarm.getActivity());
          InstanceLock lock = MutexRepository.lockInstance(bpelInstance);
          try {
            signal(execution, null, params);
          } finally {
            MutexRepository.unlockInstance(lock);
          }
        }
      }
      if (!found) {
        throw new OrchestraException("Unsupported signal " + signal);
      }
    } else {
      executeSources(execution);
    }
  }
  public boolean isCreateInstance() {
    return createInstance;
  }
  public void setCreateInstance(boolean createInstance) {
    this.createInstance = createInstance;
  }
  public List<OnMessage> getOnMessages() {
    return onMessages;
  }
  public void setOnMessages(List<OnMessage> onMessages) {
    this.onMessages = onMessages;
  }
  public List<OnAlarm> getOnAlarms() {
    return onAlarms;
  }
  public void setOnAlarms(List<OnAlarm> onAlarms) {
    this.onAlarms = onAlarms;
  }
  public List<ReceivingElement> getReceivingElements() {
    return new ArrayList<ReceivingElement>(onMessages);
  }
}
