/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.transaction.Synchronization;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.env.Transaction;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.var.Message;

public final class MessageCarrierRepository {

  private static class TransactionnalMap<K,V> implements Map<K, V> {

    private Map<K,V> mapToCopy;

    private Map<K,V> map = new HashMap<K, V>();
    private List<K> addedEntries = new ArrayList<K>();
    private List<K> removedEntries = new ArrayList<K>();
    
    public TransactionnalMap(Transaction tx, Map<K, V> mapToCopy) {
      synchronized (mapToCopy) {
        map = new HashMap<K, V>(mapToCopy);
      }
      this.mapToCopy = mapToCopy;
    }

    public void clear() {
      map.clear();
    }

    public boolean containsKey(Object key) {
      return map.containsKey(key);
    }

    public boolean containsValue(Object value) {
      return map.containsValue(value);
    }

    public Set<java.util.Map.Entry<K, V>> entrySet() {
      return Collections.unmodifiableSet(map.entrySet());
    }

    public V get(Object key) {
      return map.get(key);
    }

    public boolean isEmpty() {
      return map.isEmpty();
    }

    public Set<K> keySet() {
      return Collections.unmodifiableSet(map.keySet());
    }

    public V put(K key, V value) {
      addedEntries.add(key);
      return map.put(key, value);
    }

    public void putAll(Map< ? extends K, ? extends V> t) {
      for (Entry< ? extends K, ? extends V> e : t.entrySet()) {
        put(e.getKey(), e.getValue());
      }
    }

    public V remove(Object key) {
      if (map.containsKey(key)) {
        removedEntries.add((K) key);
      }
      return map.remove(key);
    }

    public int size() {
      return map.size();
    }

    public Collection<V> values() {
      return Collections.unmodifiableCollection(map.values());
    } 
    
    void synchronize() {
      synchronized (mapToCopy) {
        for (K key : addedEntries) {
          mapToCopy.put(key, get(key));
        }
        for (K key : removedEntries) {
          mapToCopy.remove(key);
        }
      }
    }
  }
  
  private static Map<ReplierKey, MessageCarrier> messageCarriers = 
    new HashMap<ReplierKey, MessageCarrier>();
  
  private static Map<String, MessageCarrier> pendingMessageCarriers = 
    new HashMap<String, MessageCarrier>();

  private static Map<Transaction, TransactionnalMap<ReplierKey, MessageCarrier>> msgCarriersCopies = 
    new HashMap<Transaction, TransactionnalMap<ReplierKey,MessageCarrier>>();
  private static Map<Transaction, TransactionnalMap<String, MessageCarrier>> pendingMsgCarriersCopies = 
    new HashMap<Transaction, TransactionnalMap<String,MessageCarrier>>();  
  
  private MessageCarrierRepository() {
  }
  
  private static synchronized Map<String, MessageCarrier> getPendingMessageCarriers() {
    final Transaction transaction = Environment.getCurrent().get(Transaction.class);
    if (transaction != null) {
      TransactionnalMap<String, MessageCarrier> pendingMsgCarriers = pendingMsgCarriersCopies.get(transaction);
      if (pendingMsgCarriers == null) {
        pendingMsgCarriers = new TransactionnalMap<String, MessageCarrier>(transaction, pendingMessageCarriers);
        pendingMsgCarriersCopies.put(transaction, pendingMsgCarriers);
        
        Synchronization synchro = new Synchronization() {
          public void afterCompletion(int status) {
            if (status == javax.transaction.Status.STATUS_COMMITTED) {
              pendingMsgCarriersCopies.remove(transaction).synchronize();
            }
          }
          public void beforeCompletion() {
          }
        };
        transaction.registerSynchronization(synchro);
      }
      return pendingMsgCarriers;
    } else {
      return pendingMessageCarriers;
    }
  }
  
  private static synchronized Map<ReplierKey, MessageCarrier> getMessageCarriers() {
    final Transaction transaction = Environment.getCurrent().get(Transaction.class);
    if (transaction != null) {
      TransactionnalMap<ReplierKey, MessageCarrier> msgCarriers = msgCarriersCopies.get(transaction);
      if (msgCarriers == null) {
        msgCarriers = new TransactionnalMap<ReplierKey, MessageCarrier>(transaction, messageCarriers);
        msgCarriersCopies.put(transaction, msgCarriers);
        
        Synchronization synchro = new Synchronization() {
          public void beforeCompletion() { 
          }
          public void afterCompletion(int status) {
            if (status == javax.transaction.Status.STATUS_COMMITTED) {
              msgCarriersCopies.remove(transaction).synchronize();
            }
          }
        };
        transaction.registerSynchronization(synchro);

      }
      return msgCarriers;
    } else {
      return messageCarriers;
    }
  }
  
  public static synchronized void addMessageCarrier(ReplierKey replierKey,
      MessageCarrier messageCarrier) {
    getMessageCarriers().put(replierKey, messageCarrier);
  }

  public static synchronized void associateMessage(ReplierKey replierKey,
      Message message) {
    MessageCarrier messageCarrier = getMessageCarriers().remove(replierKey);
    if (messageCarrier == null) {
      throw new OrchestraException("No MessageCarrier corresponding to key = "
          + replierKey);
    }
    messageCarrier.setMessage(message);
  }

  public static synchronized void addPendingMessageCarrier(
      MessageCarrier messageCarrier) {
    getPendingMessageCarriers().put(messageCarrier.getUuid(), messageCarrier);
  }

  public static synchronized MessageCarrier removePendingMessageCarrier(
      String messageCarrierId) {
    return getPendingMessageCarriers().remove(messageCarrierId);
  }
}
